/*
*	lcd.h : some useful defines for the Cobalt LCD driver
*
*	Copyright 1996-99, Cobalt Networks, Inc.
*
*	By: Andrew Bose	
*           Timothy Stonis (x86 version)
*           Tim Hockin
*           Adrian Sun
*           Erik Gilling
*
* NOTE: make sure that this is useable from both kernel and user space
* 
*/
#ifndef _COBALT_LCD_H
#define _COBALT_LCD_H

/* FIXME: move into C file? */
#define LCD_CHARS_PER_LINE	40
#define MAX_IDLE_TIME		120

struct lcd_display {
        unsigned long buttons;
        int size1;
        int size2;
        unsigned char line1[LCD_CHARS_PER_LINE];
        unsigned char line2[LCD_CHARS_PER_LINE];
        unsigned char cursor_address;
        unsigned char character;
        unsigned char leds;
        unsigned char *RomImage;
};

/* the set of all leds available on Cobalt systems */
#define LED_SHUTDOWN		(1 << 0)
#define LED_WEBLIGHT		(1 << 1)
#define LED_COBALTLOGO		(1 << 2)
#define LED_ETH0_TXRX           (1 << 3)
#define LED_ETH0_LINK           (1 << 4)
#define LED_ETH1_TXRX           (1 << 5)
#define LED_ETH1_LINK           (1 << 6)
#define LED_DISK0               (1 << 7)
#define LED_DISK1               (1 << 8)
#define LED_DISK2               (1 << 9)
#define LED_DISK3               (1 << 10)
#define LED_SYSFAULT            (1 << 11)
#define LED_HEART               (1 << 12)
#define LED_SPARE               (1 << 13)
#define LED_SLED0               (1 << 14)
#define LED_SLED1               (1 << 15)
#define LED_SLED2               (1 << 16)
#define LED_SLED3               (1 << 17)

/* different lcd types */
#define LCD_TYPE_UNKNOWN                     0
#define LCD_TYPE_PARALLEL                    1
#define LCD_TYPE_PARALLEL_B                  2
#define LCD_TYPE_I2C                         3

/* 
 * Function command codes for io_ctl.
 */
#define LCD_On			1
#define LCD_Off			2
#define LCD_Clear		3
#define LCD_Reset		4
#define LCD_Cursor_Left		5
#define LCD_Cursor_Right	6
#define LCD_Disp_Left		7
#define LCD_Disp_Right		8
#define LCD_Get_Cursor		9
#define LCD_Set_Cursor		10
#define LCD_Home		11
#define LCD_Read		12		
#define LCD_Write		13	
#define LCD_Cursor_Off		14
#define LCD_Cursor_On		15
#define LCD_Get_Cursor_Pos	16
#define LCD_Set_Cursor_Pos	17
#define LCD_Blink_Off		18
#define LCD_Raw_Inst		19
#define LCD_Raw_Data		20
#define LCD_Type                21

// LED controls
#define LED_Set			40	
#define LED_Bit_Set		41
#define LED_Bit_Clear		42
#define LED32_Set		43	
#define LED32_Bit_Set		44
#define LED32_Bit_Clear		45

/* button ioctls */
#define BUTTON_Read		50  

//  Button defs
#define BUTTON_Next		0x3D
#define BUTTON_Next_B		0x7E
#define BUTTON_Reset_B		0xFC
#define BUTTON_NONE_B           0xFE
#define BUTTON_Left_B           0xFA
#define BUTTON_Right_B          0xDE
#define BUTTON_Up_B             0xF6
#define BUTTON_Down_B           0xEE
#define BUTTON_Enter_B          0xBE

#define BUTTON_Power            0x01 /* hopefully, nothing will have
					this value */
#ifdef __KERNEL__

#define FP_LED_ETH0_TXRX	LED_ETH0_TXRX
#define FP_LED_ETH0_LINK	LED_ETH0_LINK
#define FP_LED_ETH1_TXRX	LED_ETH1_TXRX
#define FP_LED_ETH1_LINK	LED_ETH1_LINK
#define FP_LED_DISK0		LED_DISK0
#define FP_LED_DISK1		LED_DISK1
#define FP_LED_DISK2		LED_DISK2
#define FP_LED_DISK3		LED_DISK3
#define FP_LED_WEB 		LED_WEBLIGHT
#define FP_LED_HEART		LED_HEART
#define FP_LED_SPARE		LED_SPARE
#define FP_LED_MASK	        (FP_LED_ETH0_TXRX| FP_LED_ETH0_LINK|FP_LED_ETH1_TXRX| FP_LED_ETH1_LINK| FP_LED_DISK0| FP_LED_DISK1| FP_LED_DISK2| FP_LED_DISK3|FP_LED_WEB) 

/* some useful cmos bits */
#define CMOS_FLAG_MIN           0x0001
#define CMOS_SYSFAULT_FLAG      0x0020
#define CMOS_OOPSPANIC_FLAG     0x0040
#define CMOS_FLAG_MAX           0x0040

struct lcd_handler {
	void (*fcn)(void);
	struct lcd_handler *next, *prev;
};

extern void cobalt_lcd_disable(char *str, int *ints);
extern void lcd_led_set(const unsigned int);
extern unsigned int lcd_led_get(void);
extern int cmos_read_flag(const unsigned int);
extern void lcd_fp_led_state(const unsigned char, const unsigned int);
extern int lcd_handler_add(struct lcd_handler *);
extern void lcd_handler_del(struct lcd_handler *);
#endif

#endif
