/* 
 * include/linux/nbd.h
 * 
 * This has been munged to look much more like what /dev/loop does.
 *
 */

#ifndef LINUX_NBD_H
#define LINUX_NBD_H

#define NBD_VERSION     1

#define NBD_SET_SOCK	_IO( 0xab, 0 )
#define NBD_CLEAR_SOCK	_IO( 0xab, 1 )
#define NBD_CLEAR_QUE   _IO( 0xab, 2 )
#define NBD_SET_INFO	_IO( 0xab, 3 )
#define NBD_GET_INFO	_IO( 0xab, 4 )
#define NBD_PRINT_DEBUG _IO( 0xab, 5 )

#define NBD_COOKIE_SIZE 64

/* options, so that you can add in plug-ins for various
 * things in the future. */
#define NBD_OPTION_NONE    0
#define NBD_OPTION_GZIP    1

struct nbd_info {
	u32 nbd_server_version, nbd_client_version;
	u32 nbd_server_opts, nbd_client_opts;
	u32 nbd_tickleval;
	u32 nbd_blksize, nbd_size_blocks;
	u64 nbd_ipaddr;
	unsigned char nbd_cookie[NBD_COOKIE_SIZE];
	unsigned char reserved[8];
};

#ifdef MAJOR_NR

#include <linux/locks.h>
#include <asm/semaphore.h>

#define LOCAL_END_REQUEST

#include <linux/blk.h>

#ifdef PARANOIA
extern int requests_in;
extern int requests_out;
#endif

static void 
nbd_end_request(struct request *req)
{
	unsigned long flags;
#ifdef PARANOIA
	requests_out++;
#endif
	spin_lock_irqsave(&io_request_lock, flags);
	if (end_that_request_first( req, !req->errors, "nbd" ))
		goto out;
	end_that_request_last( req );
out:
	spin_unlock_irqrestore(&io_request_lock, flags);
}

#define MAX_NBD 128

struct nbd_device {
	int refcnt;	
	unsigned long flags;
#define NBD_READ_ONLY    (1 << 0)
#define NBD_INITIALISED  (1 << 1)
#define NBD_RUNNING      (1 << 2)
	int harderror;		/* Code of hard error			*/
	struct socket * sock;
	struct file * file; 	/* If == NULL, device is not ready, yet	*/
	int magic;		/* FIXME: not if debugging is off	*/
	int req_pending;        /* number of current pending write requests */
	struct request *head;   /* Requests are added here... */
	struct request *tail;
	struct semaphore queue_sem, read_sem, write_sem;
	struct nbd_info info;        /* info */
	unsigned long timeout;  /* timeout value */
	u16 replysize;         /* reply size */
	void *istreamp, *ostreamp;      /* for compression */
	void *zbuf;                     /* scratch compression buffer */
	int zlen;                       /* not really necessary, but useful */
};
#endif

/* This now IS in some kind of include file...	*/

/* nbd request types. the nbd protocol didn't use a different namespace 
 * from linux/fs.h, so we have some backwards compatibility gunk to 
 * deal with. the first 8 bits are used for commands. the remaining
 * bits are for various flags. */
#define NBD_OP_READ      0
#define NBD_OP_WRITE     1
#define NBD_OP_READA     2   /* not used by nbd */
#define NBD_OP_WRITEA    3   /* not used by nbd */

/* (U) means something that's not currently used */
#define NBD_OP_CONFIG    10 /* client-supported versions/options */
#define NBD_OP_DONE      11 /* (U) done! */
#define NBD_OP_TICKLE    12 /* tickle timeouts */
#define NBD_OP_FLUSH     13 /* (U) flush any outstanding requests. */
#define NBD_OP_MASK(a)   ((a) & 0xFF)

/* 24 bits for flags */
#define NBD_FLAG_ACK     (1 << 8)   /* we want an ack for this */
#define NBD_FLAG_GZIP    (1 << 9)   /* compress bit */

/* These are send over network in request/reply magic field */

#define NBD_REQUEST_MAGIC 0x25609513
#define NBD_REPLY_MAGIC 0x67446698
/* Do *not* use magics: 0x12560953 0x96744668. */

/*
 * This is packet used for communication between client and
 * server. All data are in network byte order.
 */
#define NBD_REQUEST_SIZE  28
struct nbd_request {
	u32 magic;
	u32 type;	/* == READ || == WRITE 	*/
	char handle[8];
	u64 from;
	u32 len;
}
#ifdef __GNUC__
	__attribute__ ((packed))
#endif
;

#define NBD_REPLY_SIZE_V0 16
#define NBD_REPLY_SIZE    28
struct nbd_reply {
	u32 magic;
	u32 error;		/* 0 = ok, else error	*/
	char handle[8];		/* handle you got from request	*/
        u32 type;		/* yes, we might want to set the type */
        u32 filler;             /* filler */
        u32 len;                /* len */
}
#ifdef __GNUC__
	__attribute__ ((packed))
#endif
;

#define NBD_DATA_MAX 10240
#endif
